/*-
 * Copyright (C) @BABOLO  2002 Dec 17
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 17\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: cmd_select.c,v 1.25 2008/03/30 04:25:14 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define DBSET GLOBALDBSET

int
pgoblin_perform(pgoblin_exenv *exenv, pgoblin_r *r) {
    PGresult *res;

    ifBLIN_QV4(exenv->options->flags)
        fprintf(stderr, "+pgoblin_perform %s\n", (char*)IO_CTL->text);
    if  (!IO_CTL->text) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_perform: No query");
        return(-1);
    }
    PQEXEC(res, IO_CTL->text, PGRES_COMMAND_OK, -1, R_CONN->conn);
    DBSET->clear(res);
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_perform %d\n", 0);
    return(0);
}

int
pgoblin_select(pgoblin_exenv *exenv, pgoblin_r *r) {
    char *q;

    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "+pgoblin_select %d %d %d =%s~\n"
                                      , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CCON].r
                                      , (char*)IO_CTL->text
                                      );
    q = IO_CTL->text;
    if  (!q) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_select: No query");
        return(-1);
    }
    /** IO_PQ + **/
    ;;  PQEXEC(IO_OUT->pq, q, PGRES_TUPLES_OK, -1, R_CONN->conn);
    ;;  IO_OUT->flagc = R_CONN->flags;
    ;;  IO_OUT->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
    /** IO_PQ - **/
    ifBLIN_QV4(exenv->options->flags)
        fprintf(stderr, "-pgoblin_select %s\n", IO_OUT->pq ? "OK" : "NULL");
    return(0);
}

int
pgoblin_display(pgoblin_exenv *exenv, pgoblin_r *r) {
    ssize_t ex;

    ifBLIN_QV4(exenv->options->flags)
        fprintf(stderr, "+pgoblin_display %d %d %d\n", r[PGO_COUT].r, r[PGO_CSTY].r, r[PGO_CIN].r);
    ex = STYLESET->table(IO_OUT, R_STYLE, IO_IN, 0, -2);
    if  (ex < 0) ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_display: write error");
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_display %d\n", ex);
    return((ex < 0) ? ex : 0);
}

static const char *fetchN[] =
{ "FETCH 1 FROM pgoblin"
, "FETCH 2 FROM pgoblin"
, "FETCH 3 FROM pgoblin"
, "FETCH 5 FROM pgoblin"
, "FETCH 8 FROM pgoblin"
, "FETCH 13 FROM pgoblin"
, "FETCH 21 FROM pgoblin"
, "FETCH 34 FROM pgoblin"
, "FETCH 55 FROM pgoblin"
, "FETCH 89 FROM pgoblin"
, "FETCH 144 FROM pgoblin"
, "FETCH 233 FROM pgoblin"
, "FETCH 377 FROM pgoblin"
, "FETCH 610 FROM pgoblin"
, "FETCH 987 FROM pgoblin"
, "FETCH 1597 FROM pgoblin"
, "FETCH 2584 FROM pgoblin"
, "FETCH 4181 FROM pgoblin"
, "FETCH 6765 FROM pgoblin"
, "FETCH 10946 FROM pgoblin"
, "FETCH 17711 FROM pgoblin"
, NULL
};

int
pgoblin_strsel(pgoblin_exenv *exenv, pgoblin_r *r) {
    char *q;
    size_t i;
    ssize_t k, ex;
    pgoblin_io tr;

    ifBLIN_QV4(exenv->options->flags)
        fprintf( stderr, "+pgoblin_strsel %d %d %d %d =%s~\n"
               , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CCON].r, r[PGO_CSTY].r
               , (char*)IO_CTL->text
               );
    if  (!(q = IO_CTL->text)) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_strsel: No query");
        ERROUT(EX_DATAERR, EINVAL);
    }
    tr.flagc = R_CONN->flags;
    if  (!R_CONN->intran) {
        PQEXEC(tr.pq, "BEGIN", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(tr.pq);
    }
    asprintf(&q, "DECLARE pgoblin NO SCROLL CURSOR FOR %s", q);
    if  (q == NULL) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_strsel: No memory");
        ERROUT(EX_UNAVAILABLE, ENOMEM);
    }
    PQEXEC(tr.pq, q, PGRES_COMMAND_OK, -1, R_CONN->conn);
    DBSET->clear(tr.pq);
    free(q);
    i = 0;
    PQEXEC(tr.pq, fetchN[i], PGRES_TUPLES_OK, -1, R_CONN->conn);
    k = ex = 0;

    while (DBSET->ntuples(tr.pq) > 0) {
        ex = STYLESET->table(IO_OUT, R_STYLE, &tr, k, -1);
        ifBLIN_QV4(exenv->options->flags)
            fprintf(stderr, "ex=%d aim=%d i %d => ", ex, PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE, i);
        if  (ex < 0) break;
        k += DBSET->ntuples(tr.pq);
        DBSET->clear(tr.pq);
        if  (ex > PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && i > 1) i--;
        if  (ex < PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && fetchN[i + 1]) i++;
        ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "%d\n", i);
        PQEXEC(tr.pq, fetchN[i], PGRES_TUPLES_OK, -1, R_CONN->conn)
    }
    if  (ex >= 0) ex = STYLESET->table(IO_OUT, R_STYLE, &tr, k, -2);
    DBSET->clear(tr.pq);
    PQEXEC(tr.pq, "CLOSE pgoblin", PGRES_COMMAND_OK, -1, R_CONN->conn);
    DBSET->clear(tr.pq);
    if  (!R_CONN->intran) {
        PQEXEC(tr.pq, "END", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(tr.pq);
    }
    if  (ex < 0) ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_strsel: write error");
out:
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_strsel %d\n", 0);
    return((ex < 0) ? ex : 0);
}

int
pgoblin_binsel(pgoblin_exenv *exenv, pgoblin_r *r) {
    char *q;
    size_t i;
    ssize_t k, ex;
    pgoblin_io tr;

    ifBLIN_QV4(exenv->options->flags)
        fprintf( stderr, "+pgoblin_binsel %d %d %d %d =%s~\n"
               , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CCON].r, r[PGO_CSTY].r
               , (char*)IO_CTL->text
               );
    if  (!(q = IO_CTL->text)) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_binsel: No query");
        ERROUT(EX_DATAERR, EINVAL);
    }
    tr.flagc = R_CONN->flags;
    if  (!R_CONN->intran) {
        PQEXEC(tr.pq, "BEGIN", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(tr.pq);
    }
    asprintf(&q, "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR %s", q);
    if  (q == NULL) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_binsel: No memory");
        ERROUT(EX_UNAVAILABLE, ENOMEM);
    }
    PQEXEC(tr.pq, q, PGRES_COMMAND_OK, -1, R_CONN->conn);
    DBSET->clear(tr.pq);
    free(q);
    i = 0;
    PQEXEC(tr.pq, fetchN[i], PGRES_TUPLES_OK, -1, R_CONN->conn);
    k = ex = 0;

    while (DBSET->ntuples(tr.pq) > 0) {
        ex = STYLESET->table(IO_OUT, R_STYLE, &tr, k, -1);
        ifBLIN_QV4(exenv->options->flags)
            fprintf(stderr, "ex=%d aim=%d i %d => ", ex, PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE, i);
        k += DBSET->ntuples(tr.pq);
        DBSET->clear(tr.pq);
        if  (ex < 0) break;
        if  (ex > PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && i > 0) i--;
        if  (ex < PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE && fetchN[i + 1]) i++;
        ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "%d\n", i);
        PQEXEC(tr.pq, fetchN[i], PGRES_TUPLES_OK, -1, R_CONN->conn)
    }
    if  (ex >= 0) ex = STYLESET->table( IO_OUT, R_STYLE, &tr, k, -2);
    DBSET->clear(tr.pq);
    PQEXEC(tr.pq, "CLOSE pgoblin", PGRES_COMMAND_OK, -1, R_CONN->conn);
    DBSET->clear(tr.pq);
    if  (!R_CONN->intran) {
        PQEXEC(tr.pq, "END", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(tr.pq);
    }
    if  (ex < 0) ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_binsel: write error");
out:
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_binsel %d\n", 0);
    return((ex < 0) ? ex : 0);
}

int
pgoblin_binpre(pgoblin_exenv *exenv, pgoblin_r *r) {
    PGresult *res;
    char *q;

    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "+pgoblin_binpre %d %d %d =%s~\n"
                                      , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CCON].r
                                      , (char*)IO_CTL->text
                                      );
    if  (!(q = IO_CTL->text)) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_binpre: No query");
        return(-1);
    }
    if  (!R_CONN->intran) {
/*  if  (PQtransactionStatus(R_CONN->conn) != PQTRANS_INTRANS) { */
        PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(res);
    }
    asprintf(&q, "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR %s", q);
    if  (q == NULL) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_binpre: No memory");
        return(-1);
    }
    PQEXEC(res, q, PGRES_COMMAND_OK, -1, R_CONN->conn);
    DBSET->clear(res);
    free(q);
    /** IO_PQ + **/
    ;;  PQEXEC(IO_OUT->pq, "FETCH 2147483647 FROM pgoblin", PGRES_TUPLES_OK, -1, R_CONN->conn);
    ;;  IO_OUT->flagc = R_CONN->flags;
    ;;  IO_OUT->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
    /** IO_PQ - **/
    PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1, R_CONN->conn);
    DBSET->clear(res);
    if  (!R_CONN->intran) {
        PQEXEC(res, "END", PGRES_COMMAND_OK, -1, R_CONN->conn);
        DBSET->clear(res);
    }
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_binpre %d\n", 0);
    return(0);
}
